autowatch=1;
outlets=3;

//var m.translate = new Array();

var m = new Object; //short for "MIDI" - we put both objects under one roof to make it easier to read/write into JSON.
m.translate = new Object;
m.messages = new Object;
var midichunk = new Array();

var groupon=new Task(midiout,this);
var ismaxforlive = 0;
var learning = 0;
var tolearn = "";
var rangemin=0;
var rangemax=127;
var grpindex = 0;
var btntype = 0 //0 is momentary, 1 = tog;
var learntype = 0;
var miditype = -1;
var moff = 0;
var mon = 127;

var filepath = max.apppath+"/MIDIsettings.json"; //default location

var DEBUG = 0;

function allnotesoff(){
    //error control
}

function mfl(v){
	ismaxforlive = v>0;
}

function dbug(v){
    DEBUG = v;
}

//a macro to roll learning and message all in one, so you can quickly create new contexts for the controller without loading new files.
/*
examples:
latch 144 5 enable 1 //note 5 will translate to a message "enable" and send out toggled messages
latch 145 10 oneshot 0 // note 10 on channel 2 will translate to a message "oneshot" and acts as a momentary button
latch 176 10 scrub -1 1 // cc 10 will translate to a message "scrub" and scale midi values to a range of -1.0 to 1.0
*/
function latch(){
    var a=arrayfromargs(arguments);
    var v = a[0];
    var midi = a.slice(0,2);
    var tolearn = a.slice(2);
    if(DEBUG) post("\nlatch",tolearn,midi);
    //setup the message and turn learning on with a leading "1":
    tolearn.splice(0,0,1); //add a 1 to the beginning of the array
    learn(tolearn);
    //for program and channel pressure, we don't need to adjust, 
    //but for others, we'll need to add a placeholder value for the midimatch() function
    if(v>=192 && v<=207){
        midimatch(midi[0],midi[1]);
    }else{
        midimatch(midi[0],midi[1],64);
    }
}

function update(){ //ideally, update stored values from pattrstorage
    var a = arrayfromargs(arguments);
    var checkmess = a[0];
    var thevalue = a.slice(1);
    var key,min,max,range,m1,m2,mval; 
    if(DEBUG) post("\ncheckit",checkmess,"val",thevalue);
    if(m.messages[checkmess]){
        if(m.messages[checkmess].group){
            //first turn off the group buttons:
            for(var i in m.messages[checkmess]){
                if(i!="group"){
                    var offkey = m.messages[checkmess][i];
                    b1 = offkey & 255;
                    b2 = (offkey >> 8) & 255;
                    if(DEBUG) post("\ngrp-off",checkmess,i,offkey," : ",b1,b2);
                    OFF = m.translate[offkey].rmin;
                    midiout(b1,b2,OFF);
                }
            }
            //then turn this on one:
            key=m.messages[checkmess][thevalue]
            m1 = key & 255;
            m2 = (key >> 8) & 255;
            ON = m.translate[key].rmax;
            if(DEBUG) post("\ngrp-on",checkmess,i,offkey," : ",b1,b2,ON);
            if(ismaxforlive){
				//schedule with a task, since Live doesn't seem to like this note on sent after a flurry of offs:
				groupon.cancel();
				groupon.arguments = [m1,m2,ON];
				groupon.schedule(10);
            }else{
            	midiout(m1,m2,ON);
            }
        }else{ //pots, buttons, and toggles
            key = m.messages[checkmess];
            m.translate[key].value = thevalue;
            if(m.translate[key].btype){
                m.translate[key].bval = thevalue;
                if(m.translate[key].btype==2){
                    m.translate[key].tog = thevalue; 
                }
            }
            min = m.translate[key].rmin;
            max = m.translate[key].rmax;
            range = Math.abs(max-min);
            m1 = key & 255;
            m2 = (key >> 8) & 255;
            mval = Math.floor(127*(thevalue-min)/range);
            //check for strictly defined midi out values for RGB:
            if(m.translate[key].moff!=null && thevalue==0){
                mval = m.translate[key].moff;
            }
            if(m.translate[key].mon!=null && thevalue>0){
                mval = m.translate[key].mon;
            }
            //post("\nmidi",127*thevalue,min,range,mval);
            //valout = (range*v/127)+min;
            if(DEBUG) post("\nmout",key,m1,m2,mval,"__",min,max,range);
            midiout(m1,m2,mval);
        }
    }
}

function learn(){
    btntype = 0; //reset
    learntype = 0; //reset
    grpindex = 0; //reset
    moff = null;
    mon = null;
    
//if 2 args, then first turns learning on/off, second provides a message, and range is 0-127 e.g. "learn 1 Volume"
    var a = arrayfromargs(arguments);
    //post("\nlearn length",a.length);
    //if there's just 1 argument, then we assume it is an array that came from a "latch" and we are going to have to flatten the array
     if(a.length==1){
        var tmp = a[0].slice(0); //copy contents to a temporary buffer
        a = tmp.slice(0);  //copy contents into "a" - now it has a length>1 and we can deal with it below
        if(DEBUG) post("\nlatch? new length",a.length);    
    }
    learning = a[0];
    tolearn = a[1];
    rangemin = 0;
    rangemax = 127;
    learntype = 0; //learning a range
//if 4 args, then range is 3rd and 4th e.g. "learn 1 Delay_Wet 0 1"
    if(a.length==4){ //then we have ranges to define
        rangemin = a[2];
        rangemax = a[3];
        learntype = 0; //learning a range
    }
//if 3 args, then 3rd defines button behavior e.g. "learn 1 PlayPause 1" for a toggle or "learn 1 PresetInc 0" for momentary
    if(a.length==3){ //then we have a button type (momentary or tog)
         btntype = a[2];
         learntype = 1; //a flag that indicates learning needs to account for the type of button
    }
//if 5 args, then 3rd and 4th defines on/off values and 5 defines button type (momentary or tog) 
//e.g. "learn 1 PlayPause 0 1 1" for a toggle or "learn 1 Shift 0 1 0" for momentary, 
    if(a.length==5){ //then we have a button type
        rangemin = a[2];
        rangemax = a[3];
        btntype = a[4];
        learntype = 1; //learning a button
    }
//if 6 args, then it is a group type
//or "learn 1 Clip 0 127 3 1" for group type. The "0" and "127" provide range (color), the "3" defines this as a group button, and the 1 provides the group index 
    if(a.length==6){ //this is a group type
        rangemin = a[2];
        rangemax = a[3];
        btntype = a[4];
        learntype = 1; //learning a button
        grpindex = a[5];
    }
//if 7 args, then 3rd and 4th defines on/off values and 5 defines button type (momentary or tog) and 6 & 7 define midiout values for RGB colors
//e.g. "learn 1 PlayPause 0 1 1 1 16" would make the button white for off, red for on
    if(a.length==7){ //then we have a button type
        rangemin = a[2];
        rangemax = a[3];
        btntype = a[4];
        moff = a[5];
        mon = a[6];
        learntype = 1; //learning a button
    }
    if(DEBUG) var junk = ["slider","button"];
    if(DEBUG) post("\nlearn",tolearn,rangemin,rangemax,btntype,junk[learntype]);
}


/*
m.translate[key] example:
        "17552": {
            "mess": "VideoIn",
            "rmin": 0,
            "rmax": 1,
            "btype": 2,
            "tog": true,
            "pcount": 0,
            "bval": false,
            "btnchange": false
        },
*/

function midimatch(){
    var key,value;
    var a = arrayfromargs(arguments);
    if(a.length == 3){
        key = a[0] | (a[1] << 8);
        value = a[2];
    }else{
        key = a[0];
        value = a[1];
    } 
    if(DEBUG) post("\nmatching",miditype,a.length,"--",key,value);
    if(learning){
        //the values for this were setup in the function "learn"
        if(DEBUG) post("\nlearning",tolearn,key);
        if(!m.translate[key]){ //if there's nothing at this key, make the object
           m.translate[key] = new Object;
        }else{
            if(DEBUG) post("\nremoving",m.translate[key].mess,"from m.messages",key);
            var toremove = m.translate[key].mess;
            m.messages[toremove]=null;
        }
        m.translate[key].mess=tolearn;
        m.translate[key].rmin=rangemin;
        m.translate[key].rmax=rangemax;
        if(moff!=null){
            m.translate[key].moff = moff;
        }
        if(mon!=null){
            m.translate[key].mon = mon;
        }
        if(learntype){
            //1 is momentary, 2 is tog, 3 is msg only, 4 is group. Making it >0 makes it easier later in messout()...
           m.translate[key].btype=(btntype+1); 
           m.translate[key].tog=0;
        }
        if(DEBUG) post("\nassgn",key,m.translate[key].mess,m.translate[key].rmin,m.translate[key].rmax,m.translate[key].btype)
        //make accommodations for group type:
        if(m.translate[key].btype==4){
            m.translate[key].mess=[tolearn,grpindex];
            if(!m.messages[tolearn]) {
                m.messages[tolearn] = new Object;
            }
            m.messages[tolearn].group = true; //so we can later check if this message is part of a group
            m.messages[tolearn][grpindex]=key;
        }else{
            m.messages[tolearn] = key; //so we can easily associate the message with the key        
        }
        learning = 0;
    }else if(m.translate[key]){
        messout(key,value);
    }
}

//var btnchange = -1;
//var bval; 
function messout(k,v){
    var btn,valout,msgval,isgrp;
    var key = k;
    var msgout = m.translate[key].mess;
    if(m.translate[key].btype==4){
        isgrp = 1;
        msgout = m.translate[key].mess[0];
        msgval = m.translate[key].mess[1];
    }
    var min = m.translate[key].rmin;
    var max = m.translate[key].rmax;
    var range = Math.abs(max-min);
    if(DEBUG) post("\nmsg",msgout,"out range",min,max,":",range);
    //post("\nisgrp",isgrp,msgout,msgval);
    if(m.translate[key].btype) 
        btn = m.translate[key].btype;
    if(btn){
        var bval = v>0;
        m.translate[key].bval = bval;
        if(m.translate[key].bval!=m.translate[key].btnchange){
            switch(btn){
                case 1: //momentary
                    if (range==127){
                        msgval = bval;
                    }else{
                        msgval = (1-bval)*min + (bval*max);
                    }
                    outlet(0,msgout,msgval);
                    m.translate[key].btnchange = m.translate[key].bval;
                break;
                
                case 2: //toggle
                    if(m.translate[key].bval){
                        m.translate[key].tog = 1-m.translate[key].tog;
                        outlet(0,msgout,m.translate[key].tog);
                    }
                    m.translate[key].btnchange = m.translate[key].bval;
                break;
                
                case 3: //msg only
                break;
                
                case 4: //group
                    //post("\ng",msgout,msgval);
                    outlet(0,msgout,msgval);
                break;
            }
        }
    }else{
        valout = (range*v/127)+min;
        m.translate[key].value = valout;
        outlet(0,msgout,valout);
    }
} 

//*******************************************************************************************
// parse midi data direct from "midiin" object:
var issysex = 0;
//if cc data is coming in from the controller, 
//we don't want to mirror it back to the controller - 
//it's redundant and possibly causes action problems on encoders
var midigate = 1; 
function msg_int(v){
    if(issysex==0){
    	midigate = 1; //is closed only by cc messages
        if(v>=128 && v<=143){ 
            miditype = 7; //notes off <noteID&ch,note#,vel>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=144 && v<=159){ 
            miditype = 1; //notes <noteID&ch,note#,vel>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=160 && v<=175){
            miditype = 2; //after(poly)touch <polyID&ch,note#,val>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=176 && v<=191){
            miditype = 3; //ctlr <ctlID&ch, cc#, val>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=192 && v<=207){ 
            miditype = 4; //pgm ch <pgmID&ch,val>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=208 && v<=223){
            miditype = 5; //ch. pressure <chprID&ch, val>
            counter = 0;
            midichunk = new Array();
        }
        if(v>=224 && v<=239){
            miditype = 6; //pitch bend <pbID&ch, msb, lsb>
            counter = 0;
            midichunk = new Array();
        }
    }
    if(v==240){
        issysex = 1;
    }
    if(v==247){
        issysex = 0;
    }

    switch(miditype){
        case 1: //note ON
            midichunk[counter] = v;
            if (counter==2) midimatch(midichunk[0],midichunk[1],midichunk[2]);
            counter++;
        break;
        
        case 2: //after(poly)touch
             midichunk[counter] = v;
            if (counter==2) midimatch(midichunk[0],midichunk[1],midichunk[2]);
            counter++;
        break;
        
        case 3: //cc
        	midigate = 0; //close the gate so cc data isn't mirrored back. 
            midichunk[counter] = v;
            if (counter==2) {
                midimatch(midichunk[0],midichunk[1],midichunk[2]);
                //all notes off m.messages use CC# 123
                midigate = 1; //open it back up now that we are done processing this cc.
                if(midichunk[1]==123 && midichunk[2]==0) allnotesoff();
            }
            counter++;
        break;
        
        case 4: //pgm changes
             midichunk[counter] = v;
            if (counter==1) midimatch(midichunk[0],midichunk[1]);
            counter++;
        break;
        
        case 5: //ch. pressure
             midichunk[counter] = v;
            if (counter==1) midimatch(midichunk[0],midichunk[1]);
            counter++;
        break;
        
        case 6://pitch bend
            midichunk[counter] = v;
            if (counter==2) midimatch(midichunk[0],midichunk[1],midichunk[2]);
            counter++;
        break;
        
        case 7: //note OFF - converted to noteon, vel=0 with the "+16" for matching purposes
            midichunk[counter] = v;
            if (counter==2)
                midimatch(midichunk[0]+16,midichunk[1],0);
            counter++;
        break;

    }
}

function midiout(){
    var midibytes = arrayfromargs(arguments);
    for(var i=0;i<midibytes.length;i++){
        if(midigate==1) outlet(1,midibytes[i]); //if the gate is closed, don't send MIDI out.
    }
}

//relies on function defined in separate file "JSON_parse.js" which needs to be in Max5/Cycling '74/jsextensions
function write(p){
    var tojson,fout;
    tojson = JSON.stringify(m,null,'\t'); 
    if(p) filepath = p; //otherwise, it's the max folder
    fout = new File(filepath,"write","TEXT");
    if (fout.isopen) {
        fout.eof = 0;
        fout.writeline(tojson);
        fout.close();
        post("\nMIDI settings write:",filepath);
    } else {
        post("\ncould not create MIDI json file:",filepath);
    }
}

function read(p) {
    clear();
    var memstr = "";
    var data = "";
    var maxchars = 800; //may need to increase size for longer files!
    if(p) filepath = p;
    var f = new File(filepath,"read");
    f.open();
    if (f.isopen) {
        post("\nMIDI settings read:",filepath);
        while(f.position<f.eof) {
            memstr+=f.readstring(maxchars);
        }
        f.close();
    } else {
        post("\nError reading MIDI file:",filepath);
    }
    //assign it to the object
    m = JSON.parse(memstr);
}

function clear(){
    //a container for all the information we need
    m = new Object;
    
    //lots of info about midi in midi.translate, all stored at [key]. key identifies a midi message like note 23 or cc 45.
    //using "Mix::Effect[1] 0.65" as an example, or "PlayPause 1"
    //midi.translate.key.mess is the pattr message attached to a midi message, such as "Mix::Effect[1]
    //midi.translate.key.value is the current value of a parameter, such as a float like 0.65 
    //midi.translate.key.btype is the control type, like b,m,cc,t, or s
    //midi.translate.key.bval is the button value for down (1) or up (0), derived from velocity
    //midi.translate.key.tog is the stored toggle value for toggle types, will be 1 or 0.
    //midi.translate.key.rmin is range min, like "-1.0"
    //midi.translate.key.rmax is range max, like "1.0"
    m.translate = new Object;
    
    //stores a key at [message] for outputting MIDI from incoming messages, 
    //e.g. midi.messages["Mix::Effect[1]"]["key"] = 11696 means "Mix::Effect[1]" is attached to cc 45 on ch. 1;
    m.messages = new Object;
}